import asyncio
import inspect
import os
import signal
import time
import traceback
from multiprocessing import Process
from threading import Thread
from typing import Callable

from utils.log import logger
from utils.system import is_linux, cur_pid
from utils.call import call_async
from .q import QM
from .server import init_base


def extend_args(task):
    from utils.config import get_global_config
    from utils.log.q import QLogM
    args = [QLogM.log_q, task.init, task.end, get_global_config()]
    args.extend(QM.qs())
    args.insert(0, len(args))

    args.extend(task.args)
    return args


def parse_args(args):
    e_num = args[0]
    args = args[1:]
    return args[:e_num], args[e_num:]


class MpDecorator:
    def __init__(self, func, grace=True):
        self.func = func
        self.grace = grace

    def __call__(self, *args, **kwargs):

        if self.grace:
            GracefulKiller(exit_now=True)

        e_args, args = parse_args(args)
        log_q, init, end, global_conf, *qs = e_args
        init_base(log=log_q, qs=qs, global_conf=global_conf)

        result = None
        if init:
            init()

        # import nest_asyncio
        # nest_asyncio.apply()

        func_name = self.func.__name__

        try:
            if inspect.iscoroutinefunction(self.func):
                logger.info(f"run async {func_name}")
                call_async(self.func(*args, **kwargs))
            else:
                logger.info(f"run {func_name}")
                self.func(*args, **kwargs)
        except ErrGracefulKiller:
            print(f"{cur_pid()} {func_name} grace exit")
        except Exception as e:
            print(f"{cur_pid()} {func_name} {e}")
            msg = f"{cur_pid()} {func_name} exception {traceback.format_exc()}"
            logger.error(msg)

        if end:
            end()
        return result


class Task:
    def __init__(self, handle, name, *args,
                 task_q_name: str = None,
                 init: Callable = None,
                 end: Callable = None,
                 **kwargs):
        self.handle = handle
        self.args = args
        self.name = name
        self.kwargs = kwargs
        self.init = init
        self.end = end
        self.task_q_name = task_q_name


class _MultiM:
    def __init__(self):
        self._p: dict[str, Task] = {}
        self._rp: dict[int, tuple[Task, Process]] = {}
        self._t: dict[str, Task] = {}
        self._rt: dict[int, tuple[Task, Thread]] = {}
        self._killer: GracefulKiller = None

    def add_p(self, name, handler, *args, **kwargs):
        if name in self._p:
            logger.debug(f"MultiM add_p {name} exist")
            return
        self._p[name] = Task(handler, name, *args, **kwargs)

    def add_t(self, name, handler, *args, **kwargs):
        if name in self._t:
            logger.debug(f"MultiM add_t {name} exist")
            return
        self._t[name] = Task(handler, name, *args, **kwargs)

    @property
    def close(self):
        return self._killer and self._killer.kill_now

    def _new_p(self, task) -> Process:
        f = MpDecorator(task.handle, grace=True)
        p = Process(
            target=f,
            kwargs=task.kwargs,
            args=extend_args(task),
        )
        p.start()
        self._rp[p.pid] = (task, p)

        msg = f"process {task.name} started {p.pid}"
        logger.info(msg)
        print(msg)
        return p

    def _new_t(self, task) -> Thread:
        f = task.handle
        t = Thread(
            target=f,
            name=task.name,
            kwargs=task.kwargs,
            args=task.args,
        )
        t.daemon = True
        t.start()
        self._rt[t.ident] = (task, t)

        msg = f"thread {task.name} started {t.ident}"
        logger.info(msg)
        print(msg)
        return t

    def _run_tasks(self):
        self._killer = GracefulKiller(exit_now=True, handle_child_exit=self.restart_child)
        for task in self._p.values():
            self._new_p(task)

        for task in self._t.values():
            self._new_t(task)

        msg = "server start success"
        logger.info(msg)
        print(msg)

    def _wait_p_done(self, timeout=None):
        for pid, task_p in self._rp.items():
            # print(f"process {pid} joining")
            _, p = task_p
            p.join(timeout)
            if p.is_alive():
                p.terminate()
            print(f"process {pid} joined")

    def _wait_t_done(self, timeout=None):
        for tid, task_t in self._rt.items():
            _, t = task_t
            # print(f"thread {tid} joining")
            t.join()
            print(f"thread {tid} joined")

    def _need_forever(self):
        return len(self._rp) > 0

    def start(self):
        self._run_tasks()

        if self._need_forever():
            msg = f"processes started in {cur_pid()}"
            print(msg)
            logger.info(msg)
            try:
                while True:
                    time.sleep(3)
                # asyncio.get_event_loop().run_forever()
            except ErrGracefulKiller:
                print()
                # print(f"will join child")

        self._wait_p_done(3)
        from utils.log import log
        log.close()
        self._wait_t_done()
        QM.close()

        print("MultiM start end")

    @staticmethod
    def get_exit_child_pid():
        cpid, _ = os.waitpid(-1, os.WNOHANG)
        return cpid

    def _restart_child_by_pid(self, pid):
        logger.info(f"child process-{pid} exceptionally exit")
        task_p = self._rp.get(pid)
        if not task_p:
            logger.info(f"restart child no {pid}")
            return
        task, p = task_p
        p.terminate()
        p = self._new_p(task)
        logger.info(f"restart child {p.pid}")

    def restart_child(self, sig, frame):
        if self.close:
            return
        try:
            cpid = self.get_exit_child_pid()
            if cpid == 0:
                logger.error(f"no child process was immediately available")
                return
            self._restart_child_by_pid(cpid)
        except ChildProcessError:
            logger.error(f"restart child fail")


MultiM = _MultiM()


class ErrGracefulKiller(Exception):
    pass


class GracefulKiller:
    kill_now = False

    def __init__(self, exit_now=True, handle_child_exit=None):
        if not exit_now:
            handler = self._exit_gracefully
        else:
            handler = self._exit_now
        signal.signal(signal.SIGINT, handler)
        signal.signal(signal.SIGTERM, handler)
        if is_linux() and handle_child_exit:
            signal.signal(signal.SIGCHLD, handle_child_exit)

    def _exit_now(self, sig, frame):
        self._exit_gracefully(sig, frame)
        raise ErrGracefulKiller(f'Received signal {sig} on line {frame.f_lineno} in {frame.f_code.co_filename}')

    def _exit_gracefully(self, sig, frame):
        self.kill_now = True
